import os
import time
import json
import random
from collections import deque


def update_nums(values):
    """recursively replace all numeric values in a dict with random floats"""
    for k, v in values.items():
        if type(v) == int or type(v) == float:
            values[k] = v * random.random()
        elif type(v) == dict:
            values[k] = update_nums(v)
    return values


def mutate(command):
    """Mutate a command in a type-specific way. Mutation logic is:
    - facts: Numeric facts change every run. Non-numeric facts are static.
    - reports: identity (not implemented)
    - catalogs: identity (not implemented)"""

    if 'values' in command:
        # we assume every numeric fact is volatile
        values = command['values']
        command['values'] = update_nums(values)
    return command


class Commands(object):
    """Represents an on-disk collection of commands in format:
    location/
      facts/
      reports/
      catalogs/"""

    def __init__(self, location='./samples'):
        self.location = location
        self.commands = None
        self.init_data()

    def init_data(self):
        """slurp everything up to avoid disk contention with PDB"""
        commands = {'facts': [], 'reports': [], 'catalogs': []}
        for path, _, files in os.walk(self.location):
            for f in files:
                data = json.load(open(os.path.join(path, f)))
                if 'facts' in path:
                    commands['facts'] = commands['facts'] + [data]
                elif 'reports' in path:
                    commands['reports'] = commands['reports'] + [data]
                elif 'catalogs' in path:
                    commands['catalogs'] = commands['catalogs'] + [data]
        self.commands = commands

    def __getitem__(self, k):
        return self.commands[k]

    def get(self, command_type):
        """get a random instance of a command type"""
        return random.choice(self.commands[command_type])


class CommandPipe(object):
    """A threadsafe ring buffer accessed in order of update timestamp,
    representing the command submission schedule and containing references to
    data pertinent to each simulated host. Access is throttled according to
    numhosts/runinterval. Multiple consumers are required to avoid hiccups."""

    def __init__(self, commands, numhosts, runinterval, events_per_report=4):
        """initialize a generator representing a realistic command sequence for
        the requested parameters, with mutation"""
        self.interval_sec = 60 * runinterval
        self.commands = commands
        t = 1000*time.time() + random.random()
        splay = t - float(self.interval_sec * 1000)
        self.state = deque()
        for i in range(numhosts):
            f = random.randint(0, len(commands['facts']) - 1)
            r = random.randint(0, len(commands['reports']) - 1)
            c = random.randint(0, len(commands['catalogs']) - 1)
            record = {"id": i, "updated": splay, "ref": [f, r, c]}
            splay = splay + self.interval_sec/float(numhosts)
            self.state.append(record)

    def __iter__(self):
        return self

    def next(self):
        record = self.state.popleft()
        elapsed = 1000*time.time() - record['updated']
        print(elapsed)
        if elapsed < self.interval_sec*1000:
            time.sleep(self.interval_sec - elapsed/1000)
        fidx, ridx, cidx = record['ref']
        f = self.commands['facts'][fidx]
        r = self.commands['reports'][ridx]
        c = self.commands['catalogs'][cidx]

        certname = 'host-' + str(record['id'])
        f['certname'] = certname
        r['certname'] = certname
        c['certname'] = certname

        if random.random() < 1:
            f = mutate(f)
        if random.random() < 1:
            r = mutate(r)
        if random.random() < 1:
            c = mutate(c)
        record['updated'] = time.time()*1000
        self.state.append(record)
        return f, r, c



import requests
import time
from requests.compat import urljoin
from commands import Commands
from commands import CommandPipe
import asyncio
import sys
import json
import ijson
import grequests
from datetime import datetime


DEFAULT_MUTATION_OPTS = {
    "events_per_report": 4
}

COMMAND_VERSIONS = {
    "replace_facts": 4,
    "store_report": 7,
    "replace_catalog": 9
}


DEFAULT_METRICS = {
    "replace_facts_time":
    "puppetlabs.puppetdb.storage:name=replace-facts-time",
    "queue_time":
    "puppetlabs.puppetdb.mq:name=global.queue-time",
    "command_parse_time":
    "puppetlabs.puppetdb.mq:name=global.command-parse-time",
    "queue_depth":
    "puppetlabs.puppetdb.mq:name=global.depth"
}


class PuppetDB(object):
    """ An interface to a PuppetDB instance """
    def __init__(self, host="localhost", port=8080):
        self.commands = None
        self.metrics = DEFAULT_METRICS
        self.baseurl = "http://%s:%s/" % (host, port)

    def __request(self, method, endpoint, data=None, params={}):
        if method == 'get':
            r = requests.get(urljoin(self.baseurl, endpoint), data=data)
            return r
        elif method == 'post':
            return grequests.post(urljoin(self.baseurl, endpoint),
                                  params=params, data=data)

    def query(self, query):
        """write the result of a query to an output stream"""
        url = urljoin(self.baseurl, '/pdb/query/v4')
        out = sys.stdout
        data = {'query': query}
        with requests.get(url, data=data, stream=True) as r:
            objects = ijson.items(r, 'item')
            out.write('[\n')
            d = next(objects)
        while d:
            out.write(json.dumps(d, indent=4, sort_keys=True))
            try:
                d = next(objects)
                out.write(",\n")
            except:
                out.write("]")
                break

    def submit_command(self, certname, command, data):
        """ submit a command to PuppetDB. Parameters are:
        certname: the certname of the agent associated with the command
        command: the command name ('replace_facts', 'store_report', etc)
        data: the full command body, as described in PDB's wire format docs"""

        data["producer_timestamp"] = datetime.now().isoformat()
        json_data = json.dumps(data)
        version = COMMAND_VERSIONS[command]
        return self.__request('post', '/pdb/cmd/v1', data=json_data,
                              params={"certname": certname,
                                      "command": command,
                                      "version": str(version)})

    def _simulate(self, stream, numhosts, runinterval):
        while True:
            reqs = []
            bufsize = 10
            for i in range(bufsize):
                f, r, c = stream.next()
                reqs += [self.submit_command(
                    f['certname'], 'replace_facts', f)]
            grequests.map(reqs)

    def _record_metrics(self, metrics, outputfile):
        while True:
            result = requests.post(
                urljoin(self.baseurl, '/metrics/v1/mbeans'),
                data=json.dumps(metrics),
                headers={"Content-Type": "application/json"}).json()

            with open(outputfile, 'a') as f:
                result.update({"timestamp": str(datetime.now())})
                f.write(json.dumps(result) + '\n')
            time.sleep(60)

    def prepare_sample(self):
        """ interactively prepare a simulation seed set """

    def simulate(self, numhosts, runinterval, events_per_report=4,
                 opts=DEFAULT_MUTATION_OPTS):
        """Submit simulated commands to PDB, targeting equivalence to a given
        number of hosts, run interval, and mutation rate. Record selected
        metrics in a log file on an interval.
        """
        if not self.commands:
            self.commands = Commands()
        s = CommandPipe(self.commands, numhosts, runinterval)
        t = datetime.now().isoformat()
        outputfile = 'simulation-logs-%s-%s-%s.txt' % (numhosts,
                                                       runinterval, t)
        nworkers = 8
        loop = asyncio.get_event_loop()

        async def sim():
            futures = [
                loop.run_in_executor(
                    None,
                    self._simulate,
                    s,
                    numhosts,
                    runinterval
                )
                for i in range(nworkers)
            ] + [loop.run_in_executor(None, self._record_metrics,
                                      self.metrics, outputfile)]
            return await asyncio.gather(*futures)

        loop = asyncio.get_event_loop()
        loop.run_until_complete(sim())
        loop.close()
